/*: TAVL_INS.C	TAVL Tree library    Version 3.01   6-feb-93
 *: Prototype:	TAVL_nodeptr tavl_insert(TAVL_treeptr, void *item, int replace)
 *: Returns:	Pointer to node inserted/found or NULL iff insufficient memory
 *: Purpose:	Insert item into TAVL_TREE;  "item" may be anything,
 *.		but the user-defined function "key_of()" given when tree was
 *.		initialized by "tavl_init()" must be able to read an
 *.		identifier from "item" which can be compared with other
 *.		indentifiers via the user defined function "cmp()".
 *.
 *.		If a datanode is found such that identifier(datanode) equals
 *.		identifier(item), the new data from item replaces the data
 *.		existing in datanode if & only if parameter replace == 1.
 *.
 *.!!NOTE!!	Programs that use this module must also link in the module
 *.		whose source is in the file "TAVLREBL.C".
 *
 *
 *  RELEASED TO THE PUBLIC DOMAIN
 *
 *  author:		    Bert C. Hughes
 *			    200 N.Saratoga
 *			    St.Paul, MN 55104
 *			    Compuserve 71211,577
 */

#include <stdlib.h>	   /* for prototype or macro "abs()" */
#include "tavlpriv.h"

TAVL_nodeptr tavl_insert(TAVL_treeptr tree, void *item, int replace)
		/*
		Using the user supplied (key_of) & (cmp) functions, *tree
		is searched for a node which matches *item. If a match is
		found, the new item replaces the old if & only if
		replace != 0.  If no match is found the item is inserted
		into *tree.  "tavl_insert" returns a pointer to the node
		inserted into or found in *tree. "tavl_insert" returns
		NULL if & only if it is unable to allocate memory for
		a new node.
		*/
{

    register TAVL_nodeptr p,q;
    register int cmpval = -1; /* cmpval must be initialized - if tree is */
    char junk;		      /* empty new node is left child of head */
    int side;
    TAVL_nodeptr y;

    /*	Locate insertion point for item.  "p" points to the node
	being tested, "q" follows "p" through tree.
    */

    q = &tree->head;   q->bf = LEFT;   p = Leftchild(q);

    while (p) {
	q = p;

	cmpval = (*tree->cmp)((*tree->key_of)(item),\
			      (*tree->key_of)(p->dataptr));
	if (cmpval < 0)
	    p = Leftchild(p);
	else if (cmpval > 0)
	    p = Rightchild(p);
	else {
	    if (replace) {
		void *temp = (*tree->make_item)(item);
		if (temp) {
		    (*tree->free_item)(p->dataptr);
		    p->dataptr = temp;
		}
		else p = NULL; /* insufficient dynamic memory - an error */
	    }
	    return p;
	}
    }

    /* wasn't found - create new node as child of q */

    y = (*tree->alloc)(sizeof(TAVL_NODE));

    if (y) {
	y->bf = 0;
	y->Lbit = THREAD;
	y->Rbit = THREAD;
	if ((y->dataptr = (*tree->make_item)(item)) == NULL) {
	    (*tree->dealloc)(y);
	    return NULL;	/* must be out of memory */
	}
    }
    else return NULL;		/* out of memory */

    if (cmpval < 0) {		/* connect to tree and thread it */
	y->rchild = 0;
	y->Lptr = q->Lptr;
	y->Rptr = q;
	q->Lbit = LINK;
	q->Lptr = y;
	side = LEFT;
    }
    else {
	y->rchild = 1;
	y->Rptr = q->Rptr;
	y->Lptr = q;
	q->Rbit = LINK;
	q->Rptr = y;
	side = RIGHT;
    }

    /*	Adjust balance factors on path up the tree from y to the first node
	with non-zero balance factor.  This is the node (subtree) that will
	be rebalanced; its parent's balance factor does not change.
    */

     while (q->bf == 0) {
	  q->bf = side;
	  side = (q->rchild ? RIGHT : LEFT);
	  q = parentOf(q);
     }

    /* Is tree balanced? */

    /*	SEE HISTORY.TXT
     if (abs(q->bf += side) > 1) rebalance_tavl(q,&junk);
    */

     q->bf += side;
     if (abs(q->bf) > 1) rebalance_tavl(q,&junk);

     return y;
}/*tavl_insert*/
